package org.kantega.jexmec;

/*
 * Copyright 2010 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.util.List;

/**
 * Interface representing components responsible for loading plugins from a ClassLoader.
 * The process of loading a plugin will typically involve:
 * <ul>
 *  <li>Locating the plugin classes available on the class path.</li>
 *  <li>Loading the plugin class from the provided ClassLoader</li>
 *  <li>Creating an instance of each plugin class</li>
 *  <li>Injecting any dependencies plugins may have on application services</li>
 * </ul>
 */
public interface PluginLoader<P extends Plugin> {
    /**
     * Load plugins from the provided ClassLoader, using the provided
     * {@link org.kantega.jexmec.ServiceLocator} to locate any application services
     * the plugins may depend on.
     * @param pluginClass the plugin class to load implementations of
     * @param classLoader the class loader to load plugins from.
     * @param serviceLocator the {@link org.kantega.jexmec.ServiceLocator} to look up application services from.
     * @return a list containing the loaded plugins
     */
    List<P> loadPlugins(Class<P> pluginClass, ClassLoader classLoader, ServiceLocator serviceLocator);

    /**
     * Unload any plugins previously loaded form the given ClassLoader
     * @param classLoader the ClassLoader used to load the plugins which should now be unloaded.
     */
    void unloadPlugins(ClassLoader classLoader);

    /**
     * Start the PluginLoader.
     */
    void start();

    /**
     * Stop the PluginLoader.
     */
    void stop();
}
