package org.kantega.jexmec.simple;

/*
 * Copyright 2010 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.kantega.jexmec.ServiceKey;
import org.kantega.jexmec.ServiceLocator;

import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

/**
 * Abstract implementation of {@link org.kantega.jexmec.ServiceLocator}, capable of loading 
 * services directly for a given set of ServiceKeys ({@link #AbstractServiceLocator(java.util.Set)}) or indirectly by a specifying a <i>services</i> interface
 * ({@link #AbstractServiceLocator(Class)}.
 */
public abstract class AbstractServiceLocator implements ServiceLocator {
    private final Set<ServiceKey> serviceClasses;

    /**
     * Create a ServiceLocator capable of locating services identified by the given ServiceKeys.
     * @param serviceKeys service interface classes
     */
    public AbstractServiceLocator(Set<ServiceKey> serviceKeys) {
        this.serviceClasses = new HashSet<ServiceKey>(serviceKeys);
    }

    /**
     * Create a ServiceLocator capable of locating services identified by the given inteface.
     *
     * The service interface is expected to have one getter method for each
     * service type that the application wishes to make available to the
     * application's plugins.
     *
     * An implementation of this interface might look like this:
     *
     * <pre>
     * public interface AppServices {
     *
     *   UserDao getUserDaoService();
     *
     * }
     * </pre>
     *
     * This will make the {@code UserDao} interface available as a service.
     *
     * If you have multiple services of the same type, you need to qualify
     * them by name. This is done by adding an enum parameter to the getter. The
     * enum should have one name for each service name that you wish to expose:
     *
     *
     * <pre>
     * public interface AppServices extends Services {
     *
     *   DataSource getUserDaoService(DataSourceNames names);
     *
     *   enum DataSourceNames {userDataSource, orderDataSource}
     * }
     * </pre>
     *
     * @param servicesInterface the services interface to extract service types from.
     */
    public AbstractServiceLocator(Class servicesInterface) {
        this(getServicesFromInterface(servicesInterface));

    }

    private static Set<ServiceKey> getServicesFromInterface(Class servicesInterface) {
        if (!servicesInterface.isInterface()) {
            throw new IllegalArgumentException("The class " + servicesInterface + " is not an interface");
        }

        Set<ServiceKey> serviceClasses = new HashSet<ServiceKey>();
        for (Method method : servicesInterface.getDeclaredMethods()) {
            if (method.getParameterTypes().length > 1) {
                throw new IllegalArgumentException("The service method " + servicesInterface.getName() + "." + method.getName() + " is only allowed to take a single naming enumeration as parameter. " +
                        "Please remove parameters " + Arrays.asList(method.getParameterTypes()));
            }
            Set<String> serviceNames = new HashSet<String>();

            if (method.getParameterTypes().length == 1) {
                if(!method.getParameterTypes()[0].isEnum() ) {
                    throw new IllegalArgumentException("The service method " + servicesInterface.getName() + "." + method.getName() + " a naming enumeration as a parameter. " +
                            "Please remove or fix illegal parameter " + method.getParameterTypes());
                } else {
                    Object[] constants = method.getParameterTypes()[0].getEnumConstants();
                    for(Object object : constants) {
                        if(object instanceof Enum) {
                            serviceNames.add(Enum.class.cast(object).name());
                        }

                    }
                }
            }
            if (method.getReturnType() == void.class) {
                throw new IllegalArgumentException("The service method " + servicesInterface.getName() + "." + method.getName() + " is not allowed to return void");
            }
            if (method.getReturnType().isPrimitive()) {
                throw new IllegalArgumentException("The service method " + servicesInterface.getName() + "."
                        + method.getName() + " is not allowed to return primitive types. The method returns "
                        + method.getReturnType());
            }

            if(serviceNames.size() > 0) {
                for(String name : serviceNames) {
                    serviceClasses.add(ServiceKey.by(method.getReturnType(), name));
                }
            } else {
                serviceClasses.add(ServiceKey.by(method.getReturnType()));
            }
        }
        return serviceClasses;
    }

    /**
     * {@inheritDoc}
     */

    public Set<ServiceKey> getServiceKeys() {
        return serviceClasses;
    }
}
