package org.kantega.jexmec.simple;

/*
 * Copyright 2010 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.kantega.jexmec.*;

import java.util.*;

/**
 * Default implementation of PluginManager. Tracks the ClassLoaders registered by {@link PluginClassLoaderProvider}s and uses the provided
 * {@link org.kantega.jexmec.PluginLoader} implementations with {@link org.kantega.jexmec.ServiceLocator}s to load plugins from the ClassLoaders.
 */
public class DefaultPluginManager<P extends Plugin> implements PluginManager<P> {

    private final Class<P> pluginClass;

    private final List<PluginLoader<P>> pluginLoaders = new ArrayList<PluginLoader<P>>();

    private final List<PluginClassLoaderProvider> pluginClassLoaderProviders = new ArrayList<PluginClassLoaderProvider>();

    private final List<ServiceLocator> serviceLocators = new ArrayList<ServiceLocator>();

    private final List<PluginManagerListener<P>> pluginManagerListeners = Collections.synchronizedList(new ArrayList<PluginManagerListener<P>>());

    private final PluginRegistry registry = new PluginRegistry(new CompoundServiceLocator(serviceLocators));

    private enum State {STARTING, STARTED, STOPPING, STOPPED}

    private State state = State.STOPPED;

    /**
     * Creates a DefaultPluginManager managing the given type of plugins
     *
     * @param pluginClass the class of the plugin to manage.
     */
    public DefaultPluginManager(Class<P> pluginClass) {
        this.pluginClass = pluginClass;
    }

    /**
     * Adds a {@link org.kantega.jexmec.PluginClassLoaderProvider} to the plugin manager.
     *
     * @param pluginClassLoaderProvider the provider to add
     * @return the plugin manager
     */
    public DefaultPluginManager<P> addPluginClassLoaderProvider(PluginClassLoaderProvider pluginClassLoaderProvider) {
        assertState(State.STOPPED, "adding plugin class loader provider " + pluginClassLoaderProvider);
        pluginClassLoaderProviders.add(pluginClassLoaderProvider);
        return this;
    }

    /**
     * Adds a {@link org.kantega.jexmec.PluginLoader} to the plugin manager.
     *
     * @param pluginLoader the loader to add
     * @return the plugin manager
     */
    public DefaultPluginManager<P> addPluginLoader(PluginLoader<P> pluginLoader) {
        assertState(State.STOPPED, "adding plugin loader " + pluginLoader);
        pluginLoaders.add(pluginLoader);
        return this;
    }

    /**
     * Adds a {@link org.kantega.jexmec.ServiceLocator} to the plugin manager.
     *
     * @param serviceLocator the locator to add
     * @return the plugin manager
     */
    public DefaultPluginManager<P> addServiceLocator(ServiceLocator serviceLocator) {
        assertState(State.STOPPED, "adding service locator " + serviceLocator);
        serviceLocators.add(serviceLocator);
        return this;
    }


    /**
     * Adds a single service mapping. Shortcut for calling addServiceLocator(new StaticServiceLocator(ServiceKey, Object))
     *
     * @param serviceKey the key to add a service for
     * @param service    the service implementation
     * @return the plugin manager.
     */
    public <T> DefaultPluginManager<P> addService(ServiceKey<T> serviceKey, T service) {
        return addServiceLocator(new StaticServiceLocator(serviceKey, service));
    }

    /**
     * Starts the plugin manager.
     *
     * @return the plugin manager
     */
    public synchronized DefaultPluginManager<P> start() {
        assertState(State.STOPPED, "starting plugin manager");

        state = State.STARTING;

        checkDefaults();

        startPluginLoaders();

        startPluginClassLoaderProviders();

        state = State.STARTED;

        firePluginManagerStarted();

        return this;
    }

    /**
     * Add a sensible defaults class loader provider and plugin loader if the user has not added any of his own.
     *
     */
    private void checkDefaults() {
        if (pluginLoaders.isEmpty()) {
            pluginLoaders.add(new SimpleConstructorPluginLoader<P>());
        }
        if (pluginClassLoaderProviders.isEmpty()) {
            pluginClassLoaderProviders.add(new StaticPluginClassLoaderProvider(getClass().getClassLoader()));
        }
    }

    /**
     * Stops the plugin manager.
     *
     * @return the plugin manager.
     */
    public synchronized DefaultPluginManager<P> stop() {
        assertState(State.STARTED, "stopping plugin manager");

        state = State.STOPPING;

        stopPluginClassLoaderProviders();

        stopPluginLoaders();

        state = State.STOPPED;
        return this;
    }

    /**
     *  Start all plugin classloader providers
     */
    private void startPluginClassLoaderProviders() {
        for (PluginClassLoaderProvider provider : pluginClassLoaderProviders) {
            provider.start(registry);
        }
    }

    /**
     * Starts all plugin loaders
     */
    private void startPluginLoaders() {
        for (PluginLoader<P> pluginLoader : pluginLoaders) {
            pluginLoader.start();
        }
    }

    /**
     * Stops all plugin loaders.
     */
    private void stopPluginLoaders() {
        for (PluginLoader<P> pluginLoader : pluginLoaders) {
            pluginLoader.stop();
        }
    }

    /**
     * Stops all plugin classloader providers.
     */
    private void stopPluginClassLoaderProviders() {
        for (PluginClassLoaderProvider provider : pluginClassLoaderProviders) {
            provider.stop();
        }
    }

    /**
     * {@inheritDoc}
     */
    public synchronized List<P> getPlugins() {
        assertState(State.STARTED, "getting plugins. You need to call start() *before* calling getPlugins()");
        return registry.getPlugins();
    }


    /**
     * Notify listeners that the plugin manager has started
     */
    private void firePluginManagerStarted() {
        for (PluginManagerListener listener : pluginManagerListeners) {
            listener.pluginManagerStarted();
        }
    }

    /**
     * {@inheritDoc}
     */
    public void addPluginManagerListener(PluginManagerListener<P> pluginManagerListener) {
        pluginManagerListeners.add(pluginManagerListener);
    }

    /**
     * {@inheritDoc}
     */
    public void removePluginManagerListener(PluginManagerListener<P> pluginManagerListener) {
        pluginManagerListeners.remove(pluginManagerListener);
    }

    /**
     * {@inheritDoc}
     */
    public ClassLoader getClassLoader(P plugin) {
        return registry.getClassLoader(plugin);
    }

    /**
     * Throws an IllegalStateException if the <code>actual<code> state is different from the <code>expected</code> state.
     *
     * @param expected the expected state
     * @param when     description of when this state assertion failed
     */
    private void assertState(State expected, String when) {
        if (state != expected) {
            throw new IllegalStateException("Illegal state " + state + ". Expected " + expected + " when " + when);
        }
    }

    /**
     * A registry keeping track of plugins.
     */
    private class PluginRegistry implements PluginClassLoaderRegistry {

        private final ServiceLocator serviceLocator;

        private final Map<ClassLoader, List<P>> plugins = Collections.synchronizedMap(new HashMap<ClassLoader, List<P>>());

        private final Map<P, ClassLoader> classLoaderMap = new HashMap<P, ClassLoader>();

        public PluginRegistry(ServiceLocator serviceLocator) {
            this.serviceLocator = serviceLocator;
        }

        /**
         * {@inheritDoc}
         */
        public void addClassLoader(ClassLoader loader) {
            for (PluginManagerListener listener : pluginManagerListeners) {
                listener.classLoaderAdded(loader);
            }

            final List<P> loadedPlugins = loadPluginsFromClassLoader(serviceLocator, loader);

            for (P plugin : loadedPlugins) {
                classLoaderMap.put(plugin, loader);
            }

            plugins.put(loader, loadedPlugins);
        }

        /**
         * {@inheritDoc}
         */
        public void removeClassLoader(ClassLoader loader) {
            final List<P> pluginsRemoved = plugins.get(loader);

            for (P plugin : pluginsRemoved) {
                classLoaderMap.remove(plugin);
            }

            plugins.remove(loader);

            for (PluginManagerListener listener : pluginManagerListeners) {
                listener.classLoaderRemoved(loader);
            }

            unloadPluginsFromClassLoader(loader);
        }

        /**
         * Load plugins from a classloader using all plugin loaders.
         * @param serviceLocator the service locator needed by plugin loaders
         * @param classLoader the classloader to load plugins from
         * @return a list of loaded plugins.
         */
        private List<P> loadPluginsFromClassLoader(ServiceLocator serviceLocator, ClassLoader classLoader) {
            List<P> plugins = new ArrayList<P>();

            for (PluginLoader<P> pluginLoader : pluginLoaders) {
                plugins.addAll(pluginLoader.loadPlugins(pluginClass, classLoader, serviceLocator));
            }

            return plugins;
        }

        /**
         * Tell all plugin loaders to unload plugins loaded from the given class loader.
         * @param loader the ClassLoader for which all plugins should be unloaded.
         */
        private void unloadPluginsFromClassLoader(ClassLoader loader) {
            for (PluginLoader<P> pluginLoader : pluginLoaders) {
                pluginLoader.unloadPlugins(loader);
            }
        }

        /**
         * Return currently installed plugins
         * @return a list of installed plugins
         */
        public List<P> getPlugins() {
            List<P> plugins = new ArrayList<P>();
            for (List<P> loaderPlugins : this.plugins.values()) {
                plugins.addAll(loaderPlugins);
            }
            return plugins;
        }

        /**
         * Returns the classloader that caused the loading of the given plugin.
         * @param plugin the plugin
         * @return the classloader which the plugin was loaded from
         */
        public ClassLoader getClassLoader(P plugin) {
            return classLoaderMap.get(plugin);
        }
    }

    /**
     * A {@link org.kantega.jexmec.ServiceLocator} which delegates to a collection of {@link org.kantega.jexmec.ServiceLocator}s.
     */
    private static class CompoundServiceLocator implements ServiceLocator {
        private final Collection<ServiceLocator> serviceLocators;

        private CompoundServiceLocator(Collection<ServiceLocator> serviceLocators) {
            this.serviceLocators = serviceLocators;
        }

        public Set<ServiceKey> getServiceKeys() {
            Set<ServiceKey> serviceKeys = new HashSet<ServiceKey>();
            for (ServiceLocator serviceLocator : serviceLocators) {
                serviceKeys.addAll(serviceLocator.getServiceKeys());
            }
            return serviceKeys;

        }

        public <T> T lookupService(ServiceKey<T> serviceKey) {
            for (ServiceLocator serviceLocator : serviceLocators) {
                T service = serviceLocator.lookupService(serviceKey);
                if (service != null) {
                    return service;
                }
            }
            return null;
        }
    }
}