package org.kantega.jexmec.simple;

/*
 * Copyright 2010 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.kantega.jexmec.PluginClassLoaderProvider;
import org.kantega.jexmec.PluginClassLoaderRegistry;
import org.kantega.jexmec.Plugin;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.List;

/**
 * PluginLoader for easy of development of plugins. On startup this PluginLoader will set up monitoring of
 * a list of plugin directories provided through a system property given (by default {@link #DEFAULT_PLUGIN_HOMES_PROPERTY}.
 * For each of these directories the plugin loader will add a classloader that loads resources from the resourcesDirectory
 * (default {@link #DEFAULT_RESOURCES_DIRECTORY}) and classes from the classesDirectory (default {@link #DEFAULT_CLASSES_DIRECTORY}).
 *
 * For each directory this plugin loader will monitor the classes directory for changes. When a changed file is detected the
 * stratgy will remove the old classloader and add a new one. This allows for hot redeployment of single plugins.
 */
public class HotDeployingDevelopmentPluginClassLoaderProvider implements PluginClassLoaderProvider {

    private static final String DEFAULT_PLUGIN_HOMES_PROPERTY = "jexmec.pluginHomes";
    private String pluginHomesProperty = DEFAULT_PLUGIN_HOMES_PROPERTY;
    private static final String DEFAULT_CLASSES_DIRECTORY = "target/classes";
    private String DEFAULT_RESOURCES_DIRECTORY = "src/main/resources";
    private String resourcesDirectory = DEFAULT_RESOURCES_DIRECTORY;
    private String classesDirectory = DEFAULT_CLASSES_DIRECTORY;
    private final ClassLoader parentClassLoader;
    private Thread changeDetector;
    private List<DevelopmentPluginClassLoader> currentClassLoaders = Collections.synchronizedList(new ArrayList<DevelopmentPluginClassLoader>());
    private long changeDetectorInterval = 5000;
    private volatile boolean changeDetectorStopFlag = false;
    private Class<? extends Plugin> pluginClass;
    private PluginClassLoaderRegistry registry;
    private boolean shouldScan = true;

    public HotDeployingDevelopmentPluginClassLoaderProvider(Class<? extends Plugin> pluginClass, ClassLoader parentClassLoader) {
        this.pluginClass = pluginClass;
        this.parentClassLoader = parentClassLoader;
    }

    public void setShouldScan(boolean shouldScan) {
        this.shouldScan = shouldScan;
    }

    public void start(PluginClassLoaderRegistry registry) {
        this.registry = registry;
        String homesProperty = System.getProperty(pluginHomesProperty);

        if (homesProperty != null) {
            for (String home : homesProperty.split(",;")) {
                File homeDirectory = new File(home);
                if (homeDirectory.exists()) {
                    File classes = new File(homeDirectory, classesDirectory);
                    File resources = new File(homeDirectory, resourcesDirectory);
                    if (classes.exists()) {

                        final DevelopmentPluginClassLoader loader = createClassLoader(classes, resources);
                        currentClassLoaders.add(loader);
                        registry.addClassLoader(loader);
                    }

                }
            }

        }
        if(shouldScan && currentClassLoaders.size() > 0) {
            changeDetector = new Thread(new ChangeDetector(), "Jexmec change detector ");
            changeDetector.setDaemon(true);
            changeDetector.start();
        }

    }

    Thread getChangeDetector() {
        return changeDetector;
    }

    public void setChangeDetectorInterval(long changeDetectorInterval) {
        this.changeDetectorInterval = changeDetectorInterval;
    }

    private DevelopmentPluginClassLoader createClassLoader(File classesDirectory, File resourcesDirectory) {
        return new DevelopmentPluginClassLoader(pluginClass, classesDirectory, createResourcesDirectory(resourcesDirectory));
    }

    /**
     * This method is only here for tests to subclass
     * @param resourcesDirectory
     * @return normally the given directory
     */
    File createResourcesDirectory(File resourcesDirectory) {
        return resourcesDirectory;
    }

    public void stop() {
        changeDetectorStopFlag = true;
        changeDetector.interrupt();
    }

    public void checkForChanges() {
        for (DevelopmentPluginClassLoader loader : new ArrayList<DevelopmentPluginClassLoader>(currentClassLoaders)) {
            if (loader.getLastModified() > loader.getLastCheck()) {
                currentClassLoaders.remove(loader);
                registry.removeClassLoader(loader);
                final DevelopmentPluginClassLoader newLoader = createClassLoader(loader.getClassesDirectory(), loader.getResourcesDirectory());
                registry.addClassLoader(newLoader);
                currentClassLoaders.add(newLoader);
            } else {
                loader.setLastCheck(System.currentTimeMillis());
            }
        }
    }

    private class ChangeDetector implements Runnable {

        public void run() {
            while (true) {
                if(changeDetectorStopFlag) {
                    return;
                }
                checkForChanges();
                try {
                    Thread.sleep(changeDetectorInterval);
                } catch (InterruptedException e) {

                }
            }
        }

    }

    public void setPluginHomesProperty(String pluginHomesProperty) {
        this.pluginHomesProperty = pluginHomesProperty;
    }

    private class DevelopmentPluginClassLoader extends DefaultPluginClassLoader {

        private URLClassLoader resourcesLoader;

        private long lastCheck = 0;

        private File classesDirectory;
        private File resourcesDirectory;

        private DevelopmentPluginClassLoader(Class<? extends Plugin> pluginClass, File classesDirectory, File resourcesDirectory) {
            super(pluginClass, classesDirectory, parentClassLoader);
            this.classesDirectory = classesDirectory;
            this.resourcesDirectory = resourcesDirectory;
            final List<URL> resourceUrls = new ArrayList<URL>();
            try {
                resourceUrls.add(resourcesDirectory.toURL());
            } catch (MalformedURLException e) {
                throw new RuntimeException(e);
            }

            for(int i = 1; i < getURLs().length; i++) {
                resourceUrls.add(getURLs()[i]);
            }

            resourcesLoader = new URLClassLoader(resourceUrls.toArray(new URL[resourceUrls.size()]), HotDeployingDevelopmentPluginClassLoaderProvider.this.parentClassLoader) {
                @Override
                public InputStream getResourceAsStream(String name) {
                    final URL resource = getResource(name);
                    try {
                        return resource == null ? null : resource.openStream();
                    } catch (IOException e) {
                        throw new RuntimeException(e);
                    }
                }

                @Override
                public URL getResource(String name) {
                    if (isLocalResource(name)) {
                        return super.findResource(name);
                    } else {
                        return super.getResource(name);
                    }
                }

                @Override
                public Enumeration<URL> getResources(String name) throws IOException {
                    if (isLocalResource(name)) {
                        return super.findResources(name);
                    } else {
                        return super.getResources(name);
                    }
                }

            };

            setLastCheck(getLastModified());
        }

        @Override
        public URL getResource(String name) {
            return resourcesLoader.getResource(name);
        }

        @Override
        public Enumeration<URL> getResources(String name) throws IOException {
            return resourcesLoader.getResources(name);
        }

        @Override
        public InputStream getResourceAsStream(String name) {
            return resourcesLoader.getResourceAsStream(name);
        }

        public void setLastCheck(long lastCheck) {
            this.lastCheck = lastCheck;
        }

        public long getLastCheck() {
            return lastCheck;
        }

        public long getLastModified() {
            return getLastModified(classesDirectory);
        }

        private long getLastModified(File file) {
            if (file.isFile()) {
                return file.lastModified();
            } else {
                long max = 0;
                for (File child : file.listFiles()) {
                    max = Math.max(max, getLastModified(child));
                }
                return max;
            }
        }

        public File getClassesDirectory() {
            return classesDirectory;
        }

        public File getResourcesDirectory() {
            return resourcesDirectory;
        }
    }
}


