package org.kantega.jexmec.simple;

/*
 * Copyright 2010 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.kantega.jexmec.*;

import java.util.*;
import java.lang.reflect.*;

/**
 * Implementation of {@link PluginLoader} based on simple constructor injection.
 * Plugin classes are located by reading the file {@code META-INF/services/&lt;pluginClassName&gt;/simple.txt} from the
 * ClassLoader. Any constructor arguments are resolved as services using the {@link org.kantega.jexmec.ServiceLocator}.
 *
 * @param <P> the Plugin implementation type this instance will load
 */
public class SimpleConstructorPluginLoader<P extends Plugin> implements PluginLoader<P> {

    private final ClassLocator<Class<P>> classLocator;

    public SimpleConstructorPluginLoader(ClassLocator<Class<P>> classLocator) {
        this.classLocator = classLocator;
    }

    public SimpleConstructorPluginLoader() {
        classLocator = null;
    }

    public List<P> loadPlugins(Class<P> pluginClass, ClassLoader classLoader, ServiceLocator serviceLocator) {
        ClassLocator<Class<P>> classLocator = this.classLocator != null ? this.classLocator : new DefaultClassLocator<Class<P>>(pluginClass, "simple.txt");
        final List<P> plugins = new ArrayList<P>();

        final List<Class<P>> pluginClasses = classLocator.locateClasses(classLoader);


        for (Class clazz : pluginClasses) {
            if (!pluginClass.isAssignableFrom(clazz)) {
                throw new IllegalPluginClassException("Plugin class " + clazz.getName() + " is not an instance of " + pluginClass.getName(), clazz);
            }
            if (clazz.getDeclaredConstructors().length != 1) {
                throw new MultipleConstructorsException("Plugin class " + clazz.getName() + " must have exactly one constructor", clazz);
            }
            final Constructor constructor = clazz.getDeclaredConstructors()[0];
            final Object[] params = getConstructorParameters(constructor, serviceLocator);

            try {

                final P plugin = pluginClass.cast(constructor.newInstance(params));
                plugins.add(plugin);
            } catch (InstantiationException e) {
                final String msg = "Plugin class " + clazz.getName() + " is an " + (clazz.isInterface() ? "interface" : "abstract class");
                throw new IllegalPluginClassException(msg, e, clazz);
            } catch (IllegalAccessException e) {
                throw new PluginLoadingException(e);
            } catch (InvocationTargetException e) {
                throw new PluginLoadingException(e);
            }
        }
        return plugins;
    }


    public void unloadPlugins(ClassLoader classLoader) {

    }

    public void start() {

    }

    public void stop() {
       
    }

    private Object[] getConstructorParameters(Constructor constructor, ServiceLocator serviceLocator) {
        Object[] params = new Object[constructor.getParameterTypes().length];
        for (int i = 0; i < constructor.getParameterTypes().length; i++) {
            Class<?> paramClass = constructor.getParameterTypes()[i];
            if(paramClass == NamedService.class) {
                Type[] genericParameterTypes = constructor.getGenericParameterTypes();
                ParameterizedType type = (ParameterizedType) genericParameterTypes[i];

                params[i] = new SimpleNamedService(serviceLocator, (Class) type.getActualTypeArguments()[0]);

            } else {
                if (!serviceLocator.getServiceKeys().contains(ServiceKey.by(paramClass))) {
                    throw new UnknownServiceException("Plugin  class " + constructor.getDeclaringClass() + " has an illegal constructor. Parameter " + i + " of type " + paramClass.getName() + " could not be resolved to an application service", constructor.getDeclaringClass(), paramClass);
                }
                params[i] = serviceLocator.lookupService(ServiceKey.by(paramClass));
            }
        }
        return params;
    }

    private static class SimpleNamedService implements NamedService {
        private ServiceLocator serviceLocator;
        private Class serviceClass;

        SimpleNamedService(ServiceLocator serviceLocator, Class serviceClass) {
            this.serviceLocator = serviceLocator;
            this.serviceClass = serviceClass;
        }

        public Object get(String name) {
            return serviceLocator.lookupService(ServiceKey.by(serviceClass, name));
        }
    }

}
